;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; TASK.ASM	- Multitasking routines for SimpleOS
;; (c) 2001 Breckin Loggins
;; Copy Freely
;;
BITS 32

; Import these functions
EXTERN _printReg

; Export these functions
GLOBAL _loadKernelTSS
GLOBAL _newTask

;; FUNCTION: _loadKernelTSS
;; ARGUMENTS:	None
;; PURPOSE:  Loads appropriate values into the kernel's TSS and 
;;	     loads the task register with this TSS's descriptor, 
;;	     which is put in the GDT at position 0x18

_loadKernelTSS:
	; Keep in mind that at this point you still basically own the machine
	; There are no particular restrictions on where you put things
	; Eventually this boot code will probably be moved, but while in the
	; "testing phase" you can put things pretty much where you want them
	

	; Backlink is already zero, which is good
	mov dword [tss.esp0], 0x20000	; Stack pointer
	mov word  [tss.ss0],  0x10	; Kernel Data seg
	; The other stack segs are zero, only one CPL right now
	mov dword [tss.cr3], 0
	mov dword [tss.eip], 0x0	 ; where we want to jump to
	mov dword [tss.eflags], 0x2	; flags set to default
	mov [tss.eax], eax		; Save eax
	mov [tss.ecx], ecx		; Save ecx
	mov [tss.edx]. edx		; Save edx
	mov [tss.ebx], ebx		; Save ebx
	mov dword [tss.esp], 0x20000	; Save esp
	mov [tss.ebp], ebp		; Save ebp
	mov [tss.esi], esi		; Save esi
	mov [tss.edi], edi		; Save edi
	mov word [tss.es], es		; Save es
	mov word [tss.cs], cs		; Save cs
	mov word [tss.ss], ss		; Save ss
	mov word [tss.ds], ds		; Save ds
	mov word [tss.fs], fs		; Save fs
	mov word [tss.gs], gs		; Save gs
	; No LDT, no IO map, and Trap flag not set so everything else is fine
	
	; Now set up the tss_desc
	
	mov word [tss_desc.limit], 0x68	 ; 104 Bytes
	
	; Calculate the part that goes in the bottom part of the base
	mov eax, 0x10000 + tss		 ; The whole thing
	mov word [tss_desc.base1500], ax ; Lower word
	shr eax, 0x10			 ; Get rid of the bottom 16
	mov byte [tss_desc.base2316], al ; Low byte
	mov byte [tss_desc.base3124], ah ; High byte
	mov byte [tss_desc.tdp], 0x89	 ; Type, CPL, Busy bit
	mov byte [tss_desc.lag], 0x00	 ; All these are clear
	
	; Ok we've done that, now we put that in the next free
	; space in the GDT
	cld
	mov esi, tss_desc		 ; Source
	mov edi, 0x0018			 ; Dest (next avl GDT slot)
	mov cx, 0x08			 ; 8 Bytes
	rep movsb			 ; Do it
	
		
	mov ax, 0x18
	ltr ax		           ; Load task register, this enables
				   ; Multi-processing. We don't really use this
				   ; TSS and we can't JMP to it, but now
				   ; if we create a TSS and jmp to it we should
				   ; be good to go
	
	
	add word [num_gdt_entries], 1   ; Add 1 to the number of GDT Entries
	
	
	ret			   ; Go back
	
;; FUNCTION: _newTask
;; ARGUMENTS: eax contains the linear address of the entry point
;;	      edx contains the linear address of the stack pointer
;; RETURNS:   ax  contains the selector for the task switch
;; PURPOSE:   Creates a new task to be run by the OS, put the linear address
;;	      of the code entry in eax, on return ax will contain the selector
;;	      for the TSS.  This code just creates the task, it's your job
;;	      to JMP to it by doing JMP ax:0000

_newTask:

push edx
push eax
push ebx
push ecx

; Find out where we can go
mov ecx, [num_gdt_entries]	; This is how many we have now
sub ecx, 4			; The first four are not TSS or are
				; not residing in here
xor ebx, ebx			; Zero out the base address

.continue:
	cmp ecx, 0		; Is it zero already?
	jz .through		; Yes, get out of here
	add ebx, 0x68		; 104 Bytes per entry
	dec ecx			; Minus one
	jmp .continue		; Yes bartender, I'll have another round

.through:

; Finally, we add 0x30000 to it to get our base for the tss
add ebx, 0x30000

mov eax, [esp+24]			; Get the Entry point
mov edx, [esp+20]			; Get the Stack pointer

; Set up the task segment with the info that NEEDS to be filled in
mov dword [ebx + 0x1c], 0		; Load cr3 28 bytes in
mov dword [ebx + 0x24], 0x2		; Default eflags 36 bytes in
mov dword [ebx + 0x20], eax		; EIP 32 bytes in
mov dword [ebx + 0x38], edx		; Load the stack ptr 56 bytes in
mov word  [ebx + 0x48],  es		; Load ES 72 bytes in
mov word  [ebx + 0x4c], cs		; Load CS 76 bytes in
mov word  [ebx + 0x50], ss		; Load SS 80 bytes in
mov word  [ebx + 0x54], ds		; Load DS 84 bytes in
mov word  [ebx + 0x58], fs		; Load FS 88 bytes in
mov word  [ebx + 0x5c], gs		; Load GS 92 bytes in
mov word  [ebx + 0x60], 0		; Make sure LDT is zero

; Next we add our TSS Descriptor, first get the offset to the next available
; GDT entry
xor eax, eax				; Start of GDT is zero
mov ecx, [num_gdt_entries]

.continue2:
	cmp ecx,0			; Is it zero
	jz .through2			; Yes, get out of here
	add eax, 0x08			; 8 Bytes per entry
	dec ecx				; Minus 1 on the counter
	jmp .continue2			; Around and around we go
	
.through2:

mov word [eax], 0x68			; Size (104 decimal)
mov word [eax + 2], bx			; Low part of base address
shr ebx, 0x10				; Get rid of the low part
mov byte [eax + 4], bl			; Lower high part of base addy
mov byte [eax + 5], 0x89		; Default fields for type, etc
mov byte [eax + 6], 0x00		; All clear
mov byte [eax + 7], bh			; Higher high part of base addy

; We've got our descriptor now, increment num_gdt_entries
add word [num_gdt_entries], 1

pop ecx
pop ebx
pop eax
pop edx

ret	8				; Go back, we have a new task now!!

;; Data Area

; The folowing variable keeps track of the number of GDT entries, it's 1 based
; and always includes the null selector at the beginning
num_gdt_entries		dd 3

; The fundamental task structure, the TSS
tss
	.backlink	dw 0	; A pointer to the previous TSS
	.resvd1		dw 0	; Reserved space
	.esp0		dd 0	; 32 Bits for ESP0
	.ss0		dw 0	; 16 Bits for SS0
	.resvd2		dw 0	; Reserved space
	.esp1		dd 0	; 32 Bits for ESP1
	.ss1		dw 0	; 16 Bits for SS1
	.resvd3		dw 0	; Reserved space
	.esp2		dd 0	; 32 Bits for ESP2
	.ss2		dw 0	; 16 Bits for SS2
	.resvd4		dw 0	; Reserved space
	.cr3		dd 0	; 32 Bits for CR3 (control register)
	.eip		dd 0	; 32 Bits for Instruction Pointer
	.eflags		dd 0	; 32 Bits for EFLAGS
	.eax		dd 0	; 32 Bits for EAX
	.ecx		dd 0	; 32 Bits for ECX
	.edx		dd 0	; 32 Bits for EDX
	.ebx		dd 0	; 32 Bits for EBX
	.esp		dd 0	; 32 Bits for ESP (stack pointer)
	.ebp		dd 0	; 32 Bits for EBP
	.esi		dd 0	; 32 Bits for ESI
	.edi		dd 0    ; 32 Bits for EDI
	.es		dw 0	; 16 Bits for ES
	.resvd5		dw 0	; Reserved space
	.cs		dw 0	; 16 Bits for CS
	.resvd6		dw 0	; Reserved space
	.ss		dw 0	; 16 Bits for SS (stack segment)
	.resvd7		dw 0	; Reserved space
	.ds		dw 0	; 16 Bits for DS
	.resvd8		dw 0	; Reserved space
	.fs		dw 0	; 16 Bits for FS
	.resvd9		dw 0	; Reserved space
	.gs		dw 0	; 16 Bits for GS
	.resvda		dw 0	; Reserved space
	.ldt		dw 0	; 16 Bits for Local Descriptor table
	.resvdb		dw 0    ; Reserved space
	.t		dw 0	; Only the first bit of .t is mutable
	.iomap		dw 0	; 16 Bits for the IO Map base address
tss_end		; So we can calculate how big we are (103 Bytes normally)

; As you can see, it's pretty straightforward, it just defines a structure that
; can hold the current hardware state of a segment at any given time.
; As this OS gets written, an OS specific task structure will be written to
; keep track of the software state as well

; The accompanying part, the task descriptor (goes into the GDT)
tss_desc
	.limit		dw 0	; How big is the TSS (must be > 103 dec)
	.base1500	dw 0	; Base address bits 15-00
	.base2316	db 0	; Base address bits 23-16
	.tdp		db 0	; Acronym: Type (3-0), 0 bit, DPL (6-5), P (7)
	.lag		db 0	; Acronym: Limit (3-0), AVL (4), 0 bit, 0 bit,
				;          G (7)
	.base3124	db 0	; Base address bits 31-24
tss_desc_end	; So we can calculate how big we are
